import gaiaframework.base.pipeline.pipeline
import gaiaframework.base.pipeline.artifacts
import gaiaframework.base.pipeline.forcers
import gaiaframework.base.pipeline.postprocessor
import gaiaframework.base.pipeline.predictables
import gaiaframework.base.pipeline.predictors
import gaiaframework.base.pipeline.preprocessor



import os
import mysql.connector
from fastapi import Depends, Header, HTTPException, APIRouter
from gaiaframework.base.server.cache.cache import Cache
from gaiaframework.base.server.cache.cache_utils import CacheProvider, CacheFacade

class ZIDS_Server():
    router = APIRouter()

    debug = False
    mysql_db = None
    company = None
    baseDependencies = []
    x_token = ""

    ## Initialize caching - turned off by default
    ## To enable caching change cache_type to Cache.Type.REDIS
    cache_type = Cache.Type.NONE
    cache = CacheProvider.get_cache(cache_type)
    cache_facade = CacheFacade(cache, cache_type)

    def __init__(self, x_token="fake-super-secret-token", verify_x_token=False, verify_company_token=False, debug=False):
        self.x_token = x_token
        self.debug = debug
        if verify_x_token:
            self.baseDependencies.append(Depends(self.get_token_header))

        if verify_company_token:
            self.baseDependencies.append(Depends(self.get_gaia_company))

        self.router.add_api_route("/livenessprobe", self.liveness_probe, methods=["GET"], dependencies=self.baseDependencies)
        self.router.add_api_route("/readinessprobe", self.readiness_probe, methods=["GET"], dependencies=self.baseDependencies)
        pass

    async def get_gaia_company(self, Authorization: str = Header(None), GAIA_AI_TOKEN: str = Header(None, alias="GAIA-AI-TOKEN")):
        self.company = None
        gaia_ai_token = None
        if GAIA_AI_TOKEN:
            gaia_ai_token = GAIA_AI_TOKEN
        elif Authorization:
            try:
                scheme, token = Authorization.split()
                if scheme.lower() == "bearer":
                    gaia_ai_token = token
            except:
                pass
        if not gaia_ai_token:
            raise HTTPException(status_code=401, detail="Unauthorized")
        else:
            self.get_company_by_token(gaia_ai_token)
            if not self.company:
                raise HTTPException(status_code=401, detail="Unauthorized")

    async def get_token_header(self, x_token: str = Header(...)):
        if x_token != self.x_token:
            raise HTTPException(status_code=400, detail="X-Token header invalid")

    async def verify_env(self):
        # Don't allow usage of an endpoint on production environment
        if os.environ.get('SPRING_PROFILES_ACTIVE') == 'production':
            raise HTTPException(status_code=404, detail="Endpoint not available")

    def extract_host_and_service_account(self, request):
        x_forwarded_for = request.headers.get('x-forwarded-for')
        x_goog_authenticated_user_email = request.headers.get('x-goog-authenticated-user-email')
        service_account = ''
        host = ''
        try:
            if x_goog_authenticated_user_email:
                service_account = x_goog_authenticated_user_email.replace('accounts.google.com:', '')
            if x_forwarded_for:
                host = x_forwarded_for
        except Exception as e:
            print(f'Attempt to get headers from: {request.headers} failed. Error: {e}')
            pass
        return host, service_account

    def liveness_probe(self):
        return {"alive": True}

    def readiness_probe(self):
        return {"ready": True}

    def mysql_connect(self, host, user, password, database):
        self.mysql_db = mysql.connector.connect(
            host=host,
            user=user,
            password=password,
            database=database
        )

    def run_mysql_query(self, query):
        results = None
        cursor = None
        try:
            cursor = self.mysql_db.cursor()
            cursor.execute(query)
            # Fetch the results and convert to list of JSON objects
            columns = [column[0] for column in cursor.description]
            results = []
            for row in cursor.fetchall():
                results.append(dict(zip(columns, row)))
            # json_results = json.dumps(results)
            if self.debug:
                print('results', results)
        except Exception as e:
            if self.debug:
                print('run_mysql_query', e)
        if cursor:
            cursor.close()
        return results

    def get_company_by_token(self, api_token=''):
        res = self.run_mysql_query(f"SELECT * FROM new_app_company where api_token='{api_token}'")
        if res and len(res):
            self.company = res[0]
        if self.debug:
            print('self.company', self.company)



class AsyncIterator:
    items = []
    def __init__(self, items=None):
        if items:
            self.items = items

    def add_item(self, item):
        self.items.append(item)

    async def __aiter__(self):
        for item in self.items:
            yield item


import gaiaframework.base.common
import gaiaframework.base.pipeline
import gaiaframework.base.server
import gaiaframework.base.testable
import gaiaframework.base.tester
import gaiaframework.base.trainer



from typing import Any, Dict, Generator, List, NamedTuple, Optional
import json
import requests

class StreamingGenerations():
    def __init__(self, response):
        self.response = response
        self.items = []

    def _make_response_item(self, line) -> Optional[Any]:
        if line is None:
            return None
        streaming_item = json.loads(line)
        self.items.append(streaming_item)
        return streaming_item

    def __iter__(self) -> Generator[Any, None, None]:
        if not isinstance(self.response, requests.Response):
            raise ValueError("For AsyncClient, use `async for` to iterate through the `StreamingGenerations`")

        for line in self.response.iter_lines():
            item = self._make_response_item(line)
            if item is not None:
                yield item

    async def __aiter__(self) -> Generator[Any, None, None]:
        async for line in self.response:
            item = self._make_response_item(line)
            if item is not None:
                yield item


from typing import Union, List

import logging
import sys
import time
from pythonjsonlogger import jsonlogger

root = logging.getLogger()
root.setLevel(logging.INFO)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
formatter = jsonlogger.JsonFormatter()
handler.setFormatter(formatter)
root.addHandler(handler)

logger = logging.getLogger(__name__)

import os
import json
import uvicorn
from fastapi import FastAPI, APIRouter, Depends, Header, HTTPException
from fastapi.responses import StreamingResponse
from fastapi.middleware.cors import CORSMiddleware
from starlette.requests import Request

from gaiaframework.base.server.server import ZIDS_Server

from pipeline.pipeline import generatedProjectNamePipeline
from pipeline.schema.inputs import generatedProjectNameInputs
from pipeline.schema.outputs import generatedProjectNameOutputs
from tester.test_schema.test_input import TestInputRequest
from tester.test_schema.test_output import TestOutputResponse
from tester.general_tester import generatedProjectNameGeneralTester

class OutputLogger:
    def __init__(self, endpoint) -> None:
        self.data = None
        self.host = None
        self.service_account = None
        self.endpoint = endpoint

    def save_request_info(self, data, host, service_account) -> None:
        self.data = data
        self.host = host
        self.service_account = service_account

    def log_output(self, output, retrieve_type='', pipeline_timing=None, predictable_object_count=None):

        extra_info = dict(input=self.data, output=output, from_host=self.host,
                          from_service_account=self.service_account)
        if pipeline_timing:
            extra_info.update({"pipeline_exec_time": pipeline_timing,
                               "predictable_object_count": predictable_object_count})
        logger.info(f"INFO {self.endpoint} invoked {retrieve_type}", extra_info)

## Initialized a pipeline for /predict /parse endpoints.
pipeline = generatedProjectNamePipeline()

# Initialize logger classes
predict_logger = OutputLogger('predict')

# Initialize a tester
model_tester = generatedProjectNameGeneralTester()

class Server(ZIDS_Server):

    def __init__(self):
        # self.mysql_connect(host="", user="", password="", database="")
        # verify_company_token=True will try to fetch company from gaia db using token
        # 'GAIA-AI-TOKEN' in headers or 'Authorization': f'Bearer {gaia_api_token}'
        # in order for it to work you must connect to mysql first using self.mysql_connect
        super().__init__(x_token="your-x-token", verify_x_token=False, verify_company_token=False, debug=False)
        self.router.add_api_route("/predict", self.predict, methods=["POST"], dependencies=self.baseDependencies + [])
        self.router.add_api_route("/parse", self.parse, methods=["POST"], dependencies=self.baseDependencies + [])
        self.router.add_api_route("/chat", self.chat, methods=["POST"], dependencies=self.baseDependencies + [])
        self.router.add_api_route("/test", self.test, methods=["POST"],
                                  dependencies=self.baseDependencies + [Depends(self.verify_env)],
                                  include_in_schema=False)

    def load_allowed_cors_origins(self):
        allowed_origins = []
        cors_origin_list_path = 'server/cors_allowed_origins.json'
        if os.path.exists(cors_origin_list_path):
            with open(cors_origin_list_path) as f:
                allowed_origins = json.load(f)

        return allowed_origins

    def predict(self, body: generatedProjectNameInputs, request: Request) -> List[generatedProjectNameOutputs]:
        """! Predict api endpoint, is designed to support predict projects such as
        scoops classification, valid user email etc.

        Use by post request, for example:

            headers = {"Content-Type": "application/json; charset=UTF-8", **global_headers}
            url = http://localhost:8080/predict
            data = {}  # The data to run by model.
            response = requests.post(url, json=data, headers=headers)

        """
        host, service_account = self.extract_host_and_service_account(request)
        data = body.dict()
        self.cache_facade.get_request_data(data, host, service_account)
        predict_logger.save_request_info(data, host, service_account)

        output = self.cache_facade.get('--- REPLACE WITH KEYWORD FOR FINDING KEYS INSIDE THE INPUT (data) ---')
        if len(output) > 0:
            predict_logger.log_output(output, 'cache')
            return output
        try:
            tic = time.perf_counter()
            output: List[generatedProjectNameOutputs] = pipeline.execute(**data)  # call model here
            pipeline_timing = f"{time.perf_counter() - tic:0.6f}"
            predict_logger.log_output(output,
                                      pipeline_timing=pipeline_timing,
                                      predictable_object_count=len(output) if isinstance(output, list) else 1)
            self.cache_facade.set(output)
        except Exception as ex:
            output = {'error': {'request': str(ex)}}
            logger.exception("ERROR predict invoked",
                             extra=dict(input=data, output=output, from_host=host, from_service_account=service_account))
        return output

    def parse(self, body: generatedProjectNameInputs, request: Request) -> List[generatedProjectNameOutputs]:
        '''
        parse api
        '''
        data = body.dict()
        # logger.info({"message": "parse invoked", "data": json.dumps(data)})
        # print('data', data)

        # call model here
        try:
            output: generatedProjectNameOutputs = pipeline.execute(**data)
            logger.info("INFO parse invoked", extra={"input": data, "output": output.dict()})
        except Exception as ex:
            logger.exception(msg=data, exc_info=True)
            output = {'error': {'request': ex}}
            logger.error("ERROR parse invoked", extra={"input": data, "output": output})
        # call model here
        return output
    async def chat(self, body: generatedProjectNameInputs, request: Request) -> StreamingResponse:
        '''
        parse api
        '''
        data = body.dict()
        # logger.info({"message": "parse invoked", "data": json.dumps(data)})
        # print('data', data)

        # call model here
        output = pipeline.execute_stream(**data)

        async def stream_response():
            count = 0
            final_text = ''
            async for chunk in output:
                count += 1
                final_text = chunk['text']
                yield final_text

        return StreamingResponse(stream_response())
    def test(self, input_request: TestInputRequest) -> List[TestOutputResponse]:
        """
        This is the main endpoint for testing. Every project starts with a mock response service, in order to enable
        DSP integration, but as soon as a working testing functions are ready, please replace the mock service with the
        real one.

        Parameters:
            input_request: Request as received from the DSP
        Returns:
            A list of responses, to be sent back to the DSP
        """
        response = model_tester.create_mock_response(input_request)

        """    
        model_tester.save_meta_data(input_request)
        response = []
        # Until this model will support batches, work sequentially. Once it does, pass the full list of rows.
        for request_row in input_request.rows:
            input_list = [request_row]
            tester_output = model_tester.test_batch(input_list)
            response.append(tester_output[0])
        """
        return response


## Instantiate FastAPI()
app = FastAPI()
server = Server()

## Allowed cors origins
origins = server.load_allowed_cors_origins()
## Methods allowed, for FastAPI()
methods = ["*"]
## Headers allowed, for FastAPI()
headers = ["*"]
## Credentials required, bool.
credentials = True

##
# @var allow_origins
# Cors allowed origins.
# @var allow_credentials
# Allowed credentials by FastAPI()
# @var allow_methods
# Allowed methods by FastAPI()
# @var allow_headers
# Allowed headers by FastAPI()
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=credentials,
    allow_methods=methods,
    allow_headers=headers,
)
app.include_router(server.router)

if __name__ == '__main__':
    uvicorn.run(app, host='0.0.0.0', port=8080)


from gaiaframework.base.common.component import ZIDS_Component
from gaiaframework.base.common import functions
from gaiaframework.base.common.object import ZIDS_Object
from gaiaframework.base.common.regex_handler import RegexHandler



from typing import Union, List

import logging
import sys
import time
from pythonjsonlogger import jsonlogger

root = logging.getLogger()
root.setLevel(logging.INFO)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
formatter = jsonlogger.JsonFormatter()
handler.setFormatter(formatter)
root.addHandler(handler)

logger = logging.getLogger(__name__)

class OutputLogger:
    def __init__(self, endpoint) -> None:
        self.data = None
        self.host = None
        self.service_account = None
        self.endpoint = endpoint

    def save_request_info(self, data, host, service_account) -> None:
        self.data = data
        self.host = host
        self.service_account = service_account

    def log_output(self, output, retrieve_type='', pipeline_timing=None, predictable_object_count=None):

        extra_info = dict(input=self.data, output=output, from_host=self.host,
                          from_service_account=self.service_account)
        if pipeline_timing:
            extra_info.update({"pipeline_exec_time": pipeline_timing,
                               "predictable_object_count": predictable_object_count})
        logger.info(f"INFO {self.endpoint} invoked {retrieve_type}", extra_info)